--[[
Lib for classes of Computational Geometry at IMPA

Author : Emilio Vital Brazil
Date   : 25 august 2010 

--]]

geocomp = {}

------------- eps writer

geocomp.epsLinewidth = 1
geocomp.epsPointRadius = 6
geocomp.epsScaleConst = 1
geocomp.epsMaxRes = 800
geocomp.epsBBox = { Xmin = 0 , Ymin = 0 , Xmax = geocomp.epsMaxRes , Ymax = geocomp.epsMaxRes }

function geocomp.epsDohead(  epsWindow , bBox , epsLines )
  epsWindow.Xmin = epsWindow.Xmin or geocomp.epsBBox.Xmin
  epsWindow.Ymin = epsWindow.Ymin or geocomp.epsBBox.Ymin 
  epsWindow.Xmax = epsWindow.Xmax or geocomp.epsBBox.Xmax 
  epsWindow.Ymax = epsWindow.Ymax or geocomp.epsBBox.Ymax
  
  bBox.Xmin = bBox.Xmin or geocomp.epsBBox.Xmin
  bBox.Ymin = bBox.Ymin or geocomp.epsBBox.Ymin
  bBox.Xmax = bBox.Xmax or geocomp.epsBBox.Xmax
  bBox.Ymax = bBox.Ymax or geocomp.epsBBox.Ymax

  epsLines[#epsLines + 1] = string.format( '%%!PS-Adobe-3.0 EPSF-3.0\n') 
  epsLines[#epsLines + 1] = string.format( '%%%%BoundingBox: %d %d %d %d \n', epsWindow.Xmin , epsWindow.Ymin , epsWindow.Xmax , epsWindow.Ymax) 
  epsLines[#epsLines + 1] = string.format( '%%%%Creator: geocomp.lua -- IMPA\n') 
  epsLines[#epsLines + 1] = string.format( '%%%%CreationDate: ' ) .. os.date('%c\n')
  epsLines[#epsLines + 1] = string.format( '%%%%Pages: 1\n') 
  epsLines[#epsLines + 1] = string.format( '%%%%EndComments\n') 
  epsLines[#epsLines + 1] = string.format( '%%%%EndProlog\n') 
  epsLines[#epsLines + 1] = string.format( '%%%%BeginSetup\n') 
  epsLines[#epsLines + 1] = string.format( '%%%%BeginSetup\n')
  epsLines[#epsLines + 1] = string.format( '/R %f def\n' , geocomp.epsPointRadius/geocomp.epsScaleConst ) 
  epsLines[#epsLines + 1] = string.format( '/m { moveto } bind def\n') 
  epsLines[#epsLines + 1] = string.format( '/l { lineto } bind def\n') 
  epsLines[#epsLines + 1] = string.format( '/p { R 0 360 arc closepath fill} bind def\n' ) 
  epsLines[#epsLines + 1] = string.format( '%%%%EndSetup\n\n') 
  epsLines[#epsLines + 1] = string.format( "%% data bounding box %f %f %f %f\n" , bBox.Xmin , bBox.Ymin , bBox.Xmax , bBox.Ymax )
end


function geocomp.epsDopoint( point , epsLines )
  epsLines[#epsLines + 1] = string.format( '%f %f p \n' , point.x , point.y )
  return epsString
end

function geocomp.epsDopoints( points , epsLines )
  for i = 1, #points do
    epsLines[#epsLines + 1] = string.format( '%f %f p \n' , points[i].x , points[i].y )
  end
  epsLines[#epsLines + 1] = string.format( '\n'  )
end

function geocomp.epsDopolyline( points , epsLines )  
  epsLines[#epsLines + 1] = string.format('%f %f m \n' , points[1].x , points[1].y)
  for i = 2, #points do
    epsLines[#epsLines + 1] = string.format( '%f %f l \n' , points[i].x , points[i].y )
  end
  epsLines[#epsLines + 1] = string.format('stroke\n' )
end

function geocomp.epsDopolygon( points , epsLines )
  epsLines[#epsLines + 1] = string.format('%f %f m \n' , points[1].x , points[1].y)
  for i = 2, #points do
    epsLines[#epsLines + 1] = string.format( '%f %f l \n' , points[i].x , points[i].y )
  end
  epsLines[#epsLines + 1] = string.format('closepath stroke\n')
end

function geocomp.epsDorectangle( corner , epsLines )
  epsLines[#epsLines + 1] = string.format('\n')
  epsLines[#epsLines + 1] = string.format( '%f %f m \n' , corner.bottonLeft.x , corner.bottonLeft.y )
  epsLines[#epsLines + 1] = string.format( '%f %f l \n' , corner.upperRight.x , corner.bottonLeft.y )
  epsLines[#epsLines + 1] = string.format( '%f %f l \n' , corner.upperRight.x , corner.upperRight.y )
  epsLines[#epsLines + 1] = string.format( '%f %f l \n' , corner.bottonLeft.x , corner.upperRight.y )
  epsLines[#epsLines + 1] = string.format( '%f %f l \n' , corner.bottonLeft.x , corner.bottonLeft.y )
  epsLines[#epsLines + 1] = string.format( 'closepath stroke\n' )
end

function geocomp.epsDorectangleFill( corner , epsLines )
  epsLines[#epsLines + 1] = string.format( '\n' )
  epsLines[#epsLines + 1] = string.format('%f %f m \n' , corner.bottonLeft.x , corner.bottonLeft.y )
  epsLines[#epsLines + 1] = string.format('%f %f l \n' , corner.upperRight.x , corner.bottonLeft.y )
  epsLines[#epsLines + 1] = string.format('%f %f l \n' , corner.upperRight.x , corner.upperRight.y )
  epsLines[#epsLines + 1] = string.format('%f %f l \n' , corner.bottonLeft.x , corner.upperRight.y )
  epsLines[#epsLines + 1] = string.format('%f %f l \n' , corner.bottonLeft.x , corner.bottonLeft.y )
  epsLines[#epsLines + 1] = string.format('closepath fill\n' )
end

function geocomp.epsDocircleFill( data , epsLines )
  epsLines[#epsLines + 1] = string.format('\n')
  epsLines[#epsLines + 1] = string.format( '%f %f %f 0 360 arc\n' , data.center.x , data.center.y , data.radius ) 
  epsLines[#epsLines + 1] = string.format('closepath fill\n')
end

function geocomp.epsDocircle( data , epsLines )
  epsLines[#epsLines + 1] = string.format('\n')
  epsLines[#epsLines + 1] = string.format('%f %f m \n' , data.center.x + data.radius , data.center.y )
  epsLines[#epsLines + 1] = string.format( '%f %f %f 0 360 arc\n' , data.center.x , data.center.y , data.radius ) 
  epsLines[#epsLines + 1] = string.format('closepath stroke\n' )
end

function geocomp.epsDotranslate( t , epsLines )
  tx = tx or 0
  ty = ty or 0
  epsLines[#epsLines + 1] = string.format( '%f %f translate\n' , t.x , t.y )
  return epsString
end

function geocomp.epsDoscale( s , epsLines )
  s = s or geocomp.epsScaleConst
  epsLines[#epsLines + 1] = string.format( '%f dup scale\n' , s )
end

function geocomp.epsDoSetlinewidth( lw , epsLines )
  lw = lw or geocomp.epsLinewidth
  lw = lw/(geocomp.epsScaleConst)
  epsLines[#epsLines + 1] =string.format( '%f setlinewidth \n' , lw )
end

function geocomp.epsDonothing( file ) return "" end
function geocomp.epsDowindow( file ) return "" end


function geocomp.epsDotail( epsLines )
  epsLines[#epsLines + 1] = string.format( '\n\nshowpage\n%%%%EOF\n')
end

function geocomp.epsSetHead( bBox )
  local s
  local dx = bBox.Xmax - bBox.Xmin
  local dy = bBox.Ymax - bBox.Ymin
  local Sx 
  local Sy  
  if( dx < dy ) then
    s  = geocomp.epsMaxRes / dy
    Sy = geocomp.epsMaxRes
    Sx = dx*s
  else
    s  = geocomp.epsMaxRes / dx
    Sy = dy*s
    Sx = 800
  end
  geocomp.epsScaleConst = s
  geocomp.epsBBox = { Xmin = 0 , Ymin = 0 , Xmax = Sx , Ymax = Sy }
end

----------------- get primitives

function geocomp.getwindow( file )
    local bBox = { Xmin = math.huge , Ymin = math.huge , Xmax = -math.huge , Ymax = -math.huge }
    local data = { bottonLeft , upperRight }
    data.bottonLeft = { x , y }
    data.upperRight = { x , y }
    data.bottonLeft.x, data.bottonLeft.y = file:read( "*number" , "*number" )
    data.upperRight.x, data.upperRight.y = file:read( "*number" , "*number" )
    if( data.bottonLeft.x < bBox.Xmin ) then bBox.Xmin = data.bottonLeft.x end
    if( data.bottonLeft.y < bBox.Ymin ) then bBox.Ymin = data.bottonLeft.y end
    if( data.bottonLeft.x > bBox.Xmax ) then bBox.Xmax = data.bottonLeft.x end
    if( data.bottonLeft.y > bBox.Ymax ) then bBox.Ymax = data.bottonLeft.y end
    if( data.upperRight.x < bBox.Xmin ) then bBox.Xmin = data.upperRight.x end
    if( data.upperRight.y < bBox.Ymin ) then bBox.Ymin = data.upperRight.y end
    if( data.upperRight.x > bBox.Xmax ) then bBox.Xmax = data.upperRight.x end
    if( data.upperRight.y > bBox.Ymax ) then bBox.Ymax = data.upperRight.y end
   return data , bBox
end

function geocomp.getcircle( file )
    local bBox = { Xmin = math.huge , Ymin = math.huge , Xmax = -math.huge , Ymax = -math.huge }
    local data = { center , radius }
    data.center = { x , y }
    data.center.x, data.center.y, data.radius = file:read( "*number" ,  "*number" ,  "*number" )
    if( data.center.x - data.radius < bBox.Xmin ) then bBox.Xmin = data.center.x - data.radius end
    if( data.center.y - data.radius < bBox.Ymin ) then bBox.Ymin = data.center.y - data.radius end
    if( data.center.x + data.radius > bBox.Xmax ) then bBox.Xmax = data.center.x + data.radius end
    if( data.center.y + data.radius > bBox.Ymax ) then bBox.Ymax = data.center.y + data.radius end
   return data , bBox 
end

function geocomp.getrectangle( file )
    local bBox = { Xmin = math.huge , Ymin = math.huge , Xmax = -math.huge , Ymax = -math.huge }
    local data = { bottonLeft , upperRight }
    data.bottonLeft = { x , y } 
    data.upperRight = { x , y }
    data.bottonLeft.x, data.bottonLeft.y = file:read( "*number" , "*number" )
    data.upperRight.x, data.upperRight.y = file:read( "*number" , "*number" )
    if( data.bottonLeft.x < bBox.Xmin ) then bBox.Xmin = data.bottonLeft.x end
    if( data.bottonLeft.y < bBox.Ymin ) then bBox.Ymin = data.bottonLeft.y end
    if( data.bottonLeft.x > bBox.Xmax ) then bBox.Xmax = data.bottonLeft.x end
    if( data.bottonLeft.y > bBox.Ymax ) then bBox.Ymax = data.bottonLeft.y end
    if( data.upperRight.x < bBox.Xmin ) then bBox.Xmin = data.upperRight.x end
    if( data.upperRight.y < bBox.Ymin ) then bBox.Ymin = data.upperRight.y end
    if( data.upperRight.x > bBox.Xmax ) then bBox.Xmax = data.upperRight.x end
    if( data.upperRight.y > bBox.Ymax ) then bBox.Ymax = data.upperRight.y end
   return data , bBox 
end

function geocomp.getpoints( file )
    local bBox = { Xmin = math.huge , Ymin = math.huge , Xmax = -math.huge , Ymax = -math.huge }
    local data = {}
    local N = file:read( "*number" )
    for i = 1, N do
        p = { x , y }
        p.x , p.y  = file:read( "*number" ,  "*number" )
        data[i] = p
        if( p.x < bBox.Xmin ) then bBox.Xmin = p.x end
        if( p.y < bBox.Ymin ) then bBox.Ymin = p.y end
        if( p.x > bBox.Xmax ) then bBox.Xmax = p.x end
        if( p.y > bBox.Ymax ) then bBox.Ymax = p.y end 
    end
   return data , bBox 
end

function geocomp.getpolyline( file ) return geocomp.getpoints(file) end

function geocomp.getpolygon( file ) return geocomp.getpoints(file) end

function geocomp.fileGetObjc( file )
  local st = file:read()
  if( st == nil ) then return st end 
  local tp = string.match(st,"%a+")
  if( tp == nil ) then return "nothing" end
  local funcName = "get" .. tp
  if( geocomp[funcName] == nil ) then return "nothing" end
  local data = {}
  local bBox = {}
    
  data , bBox = geocomp[funcName](file)

  return tp , data , bBox
end


function geocomp.convertEps( fileIn , fileOut )
  fileOut = fileOut or "out.eps"
  local inFile = io.stdin
  if( fileIn ) then inFile = assert( io.open(fileIn,"r") ) end
  local bBox = { Xmin = math.huge , Ymin = math.huge , Xmax = -math.huge , Ymax = -math.huge }
  local dataWindow = { Xmin , Ymin , Xmax  , Ymax }  
  local hasWindow  = false
  local epsBody = "\n"
  local tableLineBody = {}  
  local tableLineHead = {}
  while true do
    tp , data , bBoxTmp = geocomp.fileGetObjc( inFile  )
    if ( tp == nil ) then break end
    local funcName = "epsDo" .. tp
    if( geocomp[funcName] ~= nil ) then geocomp[funcName]( data , tableLineBody ) end
    if( not tp ~= "window" and tp ~= "nothing" ) then 
      if( bBoxTmp.Xmin < bBox.Xmin ) then bBox.Xmin = bBoxTmp.Xmin end
      if( bBoxTmp.Ymin < bBox.Ymin ) then bBox.Ymin = bBoxTmp.Ymin end
      if( bBoxTmp.Xmax > bBox.Xmax ) then bBox.Xmax = bBoxTmp.Xmax end
      if( bBoxTmp.Ymax > bBox.Ymax ) then bBox.Ymax = bBoxTmp.Ymax end 
    end
    if ( tp == "window" ) then
      dataWindow.Xmin = bBoxTmp.Xmin 
      dataWindow.Ymin = bBoxTmp.Ymin 
      dataWindow.Xmax = bBoxTmp.Xmax 
      dataWindow.Ymax = bBoxTmp.Ymax
      hasWindow = true
    end
  end
  geocomp.epsDotail( tableLineBody )

  if( not hasWindow ) then
    local dx = (bBox.Xmax - bBox.Xmin)*0.02
    local dy = (bBox.Ymax - bBox.Ymin)*0.02
    dataWindow.Xmin = bBox.Xmin - dx
    dataWindow.Ymin = bBox.Ymin - dy
    dataWindow.Xmax = bBox.Xmax + dx
    dataWindow.Ymax = bBox.Ymax + dy
  end

  geocomp.epsSetHead( dataWindow )
  local epsHead = geocomp.epsDohead( {} , bBox , tableLineHead )
  geocomp.epsDoscale( geocomp.epsScaleConst , tableLineHead)
  geocomp.epsDotranslate( {x = -dataWindow.Xmin , y = -dataWindow.Ymin} , tableLineHead )
  geocomp.epsDoSetlinewidth( geocomp.epsLinewidth , tableLineHead )

  local outFile = assert( io.open(fileOut,"w") )
  for i = 1 , #tableLineHead do
    outFile:write( tableLineHead[i] )
  end 
  for i = 1 , #tableLineBody do
    outFile:write( tableLineBody[i] )
  end 
end

return geocomp


